"""Data Model for Subjects in MIMIC-III and MIMIC-IV"""

import functools
from collections.abc import Callable, Iterator
from functools import cached_property
from typing import ClassVar, Self

import jax.numpy as jnp
import numpy as np
import pandas as pd
import scipy

from ._literals import NumericalTypeHint
from .base import AbstractConfig, AbstractVxData
from .coding_scheme import CodesVector, NumericScheme
from .utils import Array, np_module


class InpatientObservables(AbstractVxData):
    """
    Vectorized representation of inpatient observables.

    Attributes:
        time (Array): array of time values.
        value (Array): array of observable values.
        mask (Array): array indicating missing values.

    Methods:
        empty(size: int): creates an empty instance of InpatientObservables.
        __len__(): returns the length of the time array.
        as_dataframe(scheme: AbstractScheme, filter_missing_columns=False): converts the observables
            to a pandas DataFrame.
        groupby_code(index2code: dict[int, str]): groups the observables by code.
        segment(t_sep: Array): splits the observables into segments based on time values.
        concat(observables: Union[InpatientObservables, list[InpatientObservables]]): concatenates multiple
            instances of InpatientObservables.
        time_binning(hours: float): bins the time-series into time-windows and averages the values in each window.
    """

    time: Array  # (time,)
    value: Array  # (time, size, n_layers)
    mask: Array  # (time, size)

    def __init__(self, time: Array, value: Array, mask: Array):
        self.time = time
        self.value = value if value.ndim == 3 else value.reshape(value.shape + (1,))
        self.mask = mask

    def __check_init__(self):
        xnp = np_module(self.time)
        assert self.time.dtype == np.float64, f"Expected time to be of type float64, got {self.time.dtype}"

        assert self.mask.dtype == bool, f"Expected mask to be of type bool, got {self.mask.dtype}"

        assert self.value.ndim == 3, (
            f"Expected value to be 3D (n_timestamps, obs_dim, n_channels) or more, got {self.value.ndim}"
        )
        assert self.mask.ndim == 2, f"Expected mask to be 2D, got {self.mask.ndim}"
        assert self.value.shape[:2] == self.mask.shape, (
            f"Expected value.shape to be {self.mask.shape}, got {self.value.shape}"
        )
        assert self.time.ndim == 1, f"Expected time to be 1D, got {self.time.ndim}"
        assert len(self.time) == len(xnp.unique(self.time)), "Time stamps are not unique."
        assert len(self.time) < 2 or (self.time[:-1] < self.time[1:]).all(), "Time stamps are not sorted."

    @classmethod
    def empty(
        cls,
        size: int,
        time_dtype: type | str = np.float64,
        value_dtype: type | str = np.float16,
        mask_dtype: type | str = bool,
    ) -> Self:
        """
        Create an empty InpatientObservables object.

        Parameters:
        - size (int): the size of the InpatientObservables object.

        Returns:
        - InpatientObservables: an empty InpatientObservables object with zero time, value, and mask arrays.
        """
        return cls(
            time=np.zeros(shape=0, dtype=time_dtype),
            value=np.zeros(shape=(0, size, 1), dtype=value_dtype),
            mask=np.zeros(shape=(0, size), dtype=mask_dtype),
        )

    def __len__(self):
        """
        Returns the length of the 'time' attribute.
        """
        return self.time.shape[0]

    @cached_property
    def count(self) -> int:
        """
        Returns the number of non-missing values in the 'value' attribute.
        """
        return np.sum(self.mask)

    @classmethod
    def concat(cls, observables: list[Self] | tuple[Self, ...]) -> Self:
        """
        Concatenates a list of InpatientObservables into a single InpatientObservables object.

        Args:
            observables (Union[InpatientObservables, list[InpatientObservables]]): the list
                of InpatientObservables to concatenate.

        Returns:
            InpatientObservables: the concatenated InpatientObservables object.
        """

        if len(observables) == 0:
            return cls.empty(0)
        if all(len(o) == 0 for o in observables):
            return observables[0]
        observables = [o for o in observables if len(o) > 0]

        xnp = np_module(observables[0].time)
        time = xnp.concatenate([o.time for o in observables], axis=0)
        mask = xnp.concatenate([o.mask for o in observables], axis=0)
        value = xnp.concatenate([o.value for o in observables], axis=0)
        return cls(time=time, value=value, mask=mask)

    @staticmethod
    def _binary_agg(x: np.ndarray, m: np.ndarray) -> np.ndarray:
        return np.any(x, axis=0, keepdims=True, where=m.reshape(-1, 1, 1)).astype(x.dtype)

    @staticmethod
    def _ordinal_agg(x: np.ndarray, m: np.ndarray) -> np.ndarray:
        return np.max(x[m], axis=0, keepdims=True).astype(x.dtype)

    @staticmethod
    def _categorical_agg(x: np.ndarray, m: np.ndarray) -> np.ndarray:
        # Most frequent value (mode) for categorical.
        return scipy.stats.mode(x[m], axis=0, keepdims=True)[0].astype(x.dtype)

    @staticmethod
    def _continuous_agg(x: np.ndarray, m: np.ndarray) -> np.ndarray:
        return np.mean(x, axis=0, keepdims=True, where=m.reshape(-1, 1, 1)).astype(x.dtype)

    @staticmethod
    def agg(t: NumericalTypeHint, x: np.ndarray, m: np.ndarray) -> np.ndarray:
        match t:
            case "B":
                return InpatientObservables._binary_agg(x, m)
            case "O":
                return InpatientObservables._ordinal_agg(x, m)
            case "C":
                return InpatientObservables._categorical_agg(x, m)
            case "N":
                return InpatientObservables._continuous_agg(x, m)
            case _:
                raise ValueError(f"Unrecognized aggregation type: {t}.")

    @staticmethod
    def _time_binning_aggregate(
        x: np.ndarray, mask: np.ndarray, types: tuple[NumericalTypeHint, ...]
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Aggregates the values in a given array based on the type hint.

        Args:
            x (Array): The input array.
            type_hint (npt.NDArray[NumericaltypeHint]): The type hint.

        Returns:
            Array: The aggregated array.
        """
        assert mask.dtype == bool
        assert x.ndim == 3 and mask.ndim == 2, f"Expected x to be 3D, mask to be 2D, got ({x.ndim}, {mask.ndim})"
        assert x.shape[:2] == mask.shape, f"Expected x.shape to be {mask.shape}, got {x.shape}"
        assert x.shape[1] == len(types), f"Expected x.shape[1] to be {len(types)}, got {x.shape[1]}"

        dim_mask = mask.sum(axis=0).astype(mask.dtype)
        if not dim_mask.any():
            v = np.array([0] * np.prod(x.shape[1:]), dtype=x.dtype).reshape(1, *x.shape[1:])
            return v, dim_mask

        def _apply(ti: NumericalTypeHint, xi: np.ndarray, mi: np.ndarray, i: int) -> np.ndarray:
            if dim_mask[i] == 0:
                return np.array([0], dtype=x.dtype).reshape(1, 1, *x.shape[2:])
            return InpatientObservables.agg(ti, xi, mi)

        v = np.concatenate([_apply(ti, x[:, (dim,), :], mask[:, dim], dim) for dim, ti in enumerate(types)], axis=1)
        return v, dim_mask

    def time_binning(self, hours: float, type_hint: tuple[NumericalTypeHint, ...]) -> Self:
        """
        Bin the time-series into time-windows of length `hours`.
        The values are aggregated in each window and assigned to the
        end of the window. Aggregation by default is `mean`, except for observables at binary_indices,
        where the aggregation is `any`.

        Args:
            hours (float): length of the time-windows in hours.
            type_hint (npt.NDArray[NumericaltypeHint]): type hints for the observables vector as
                (B) binary, (O) ordinal, (C) categorical, or (N) numerical.
        Returns:
            InpatientObservables: A new instance of InpatientObservables
                with the time-series binned according to the specified
                time-windows.
        """

        if len(self) == 0:
            return self
        assert self.value.ndim == 3, (
            f"Expected value to be with shape (n_time, obs_dims, n_channels), got {self.value.shape}"
        )

        last_ts = (int(self.time[-1] / hours) + 1) * hours
        new_time = np.arange(0, last_ts + hours, hours) * 1.0
        values = []
        masks = []
        for ti, tf in zip(new_time[:-1], new_time[1:]):
            time_mask = (ti <= self.time) & (self.time < tf)
            value, mask = self._time_binning_aggregate(self.value[time_mask], self.mask[time_mask], type_hint)  # type: ignore
            values.append(value)
            masks.append(mask)

        values = np.concatenate(values, axis=0)
        masks = np.stack(masks, axis=0)

        return type(self)(time=new_time[1:], value=values, mask=masks)

    @cached_property
    def timestamps(self) -> tuple[float]:
        return tuple(self.time.tolist())

    @cached_property
    def time2index(self) -> dict[float, int]:
        return {t: i for i, t in enumerate(self.timestamps)}

    def __iter__(self) -> Iterator[tuple[float, Array, Array]]:
        return zip(self.time, self.value, self.mask)


class SegmentedInpatientObservables(InpatientObservables):
    indexed_split: Array

    # The reason for storing the index split arrays without concretely splitting the observables is to avoid
    # unnecessary computation when the observables are not accessed, inefficient hardly compressed storage of
    # sharded arrays with the accompanied overhead with each sharded item.

    def __init__(self, time: Array, value: Array, mask: Array, indexed_split: Array):
        super().__init__(time=time, value=value, mask=mask)
        self.indexed_split = indexed_split

    @classmethod
    def from_observables(cls, observables: InpatientObservables, time_split: Array) -> Self:
        return cls(
            time=observables.time,
            value=observables.value,
            mask=observables.mask,
            indexed_split=cls.indexed_split_array(observables.time, time_split),
        )

    @staticmethod
    def indexed_split_array(time: Array, time_split: Array) -> Array:
        """
        Generate split indices from the time splits, which will be used to temporally split the InpatientObservables
            arrays.

        Args:
            time (Array): array of time points of the InpatientObservables object.
            time_split (Array): array of time points used to split the InpatientObservables object.

        Returns:
            Array: array of indices used to split the InpatientObservables object.
        """
        return np.searchsorted(time, time_split)

    @property
    def segments(self) -> tuple[InpatientObservables, ...]:
        if len(self.indexed_split) == 0:
            return (InpatientObservables(self.time, self.value, self.mask),)

        time = np.split(self.time, self.indexed_split, axis=0)
        value = np.split(self.value, self.indexed_split, axis=0)
        mask = np.split(self.mask, self.indexed_split, axis=0)
        return tuple(InpatientObservables(t, v, m) for t, v, m in zip(time, value, mask))

    @property
    def n_segments(self) -> int:
        return len(self.indexed_split) + 1

    def __getitem__(self, item: int) -> InpatientObservables:
        return self.segments[item]

    def __iter__(self):
        return iter((t, v, m) for t, v, m in zip(self.time, self.value, self.mask))


class LeadingObservableExtractorConfig(AbstractConfig):
    """
    AbstractConfig for LeadingObservableExtractor.

    Attributes:
        observable_code (str): the observable code to extract the leading window from.
        leading_hours (list[float]): list of leading hours to extract. Must be sorted.
        recovery_window (float): time window in hours to mask out between a nonzero value and zero.
        entry_neglect_window (float): hours to mask out in the beginning.
        minimum_acquisitions (int): minimum number of acquisitions before extracting the leading observable.
        scheme (str): name of the observation coding scheme to use.
    """

    observable_code: str
    scheme: str
    leading_hours: list[float]
    entry_neglect_window: float
    minimum_acquisitions: int  # minimum number of acquisitions to consider
    recovery_window: float

    def __init__(
        self,
        observable_code: str,
        scheme: str,
        leading_hours: list[float] | tuple[float, ...],
        entry_neglect_window: float,
        minimum_acquisitions: int,
        recovery_window: float = 0.0,
    ):
        self.observable_code = observable_code
        self.scheme = scheme
        self.leading_hours = list(leading_hours)
        self.entry_neglect_window = entry_neglect_window
        self.minimum_acquisitions = minimum_acquisitions
        self.recovery_window = recovery_window

    def __check_init__(self):
        # `leading_hours` must be sorted.
        assert all(x <= y for x, y in zip(self.leading_hours[:-1], self.leading_hours[1:])), (
            "leading_hours must be sorted"
        )
        assert isinstance(self.scheme, str), f"Expected scheme to be a string, got {type(self.scheme)}"


class LeadingObservableExtractor(AbstractVxData):
    config: LeadingObservableExtractorConfig
    observable_scheme: NumericScheme

    def __init__(self, config: LeadingObservableExtractorConfig, observable_scheme: NumericScheme):
        self.config = config
        self.observable_scheme = observable_scheme

    def __check_init__(self):
        assert self.type_hint in ("B", "O"), (
            f"LeadingObservableExtractor only supports binary and ordinal observables, "
            f"got {self.type_hint}. Categorical and Numeric types "
            "would require custom aggregation function specific to the observation of interest,"
            "e.g. the mode of categorical or the mean of numerical. In other cases, it could be more "
            "relevant to use max/min aggregation over numeric observables. Create a feature request "
            "if you need this feature."
        )

    @cached_property
    def code_index(self) -> int:
        """
        Returns the index of the observable code.

        Returns:
            int: the index of the observable code.
        """
        return self.observable_scheme.index[self.config.observable_code]

    @cached_property
    def type_hint(self) -> NumericalTypeHint:
        """
        Returns the type hint for the observable.

        Returns:
            NumericaltypeHint: the type hint for the observable.
        """
        return self.observable_scheme.type_hint[self.config.observable_code]

    @cached_property
    def aggregation_name(self) -> str:
        """
        Returns the aggregation function based on the aggregation scheme.

        Returns:
            Callable: the aggregation function.
        """
        if self.type_hint == "B":
            return "any"
        elif self.type_hint == "O":
            return "max"
        else:
            assert False, f"unsupported type hint {self.type_hint}"

    def __len__(self):
        """
        Returns the number of leading hours.

        Returns:
            int: the number of leading hours.
        """
        return len(self.config.leading_hours)

    @cached_property
    def index2code(self):
        """
        Returns the mapping of index to code.

        Returns:
            dict[int, str]: the mapping of index to code.
        """
        desc = self.observable_scheme.desc[self.config.observable_code]
        return dict(
            zip(range(len(self.config.leading_hours)), [f"{desc}_next_{h}hrs" for h in self.config.leading_hours])
        )

    @cached_property
    def index2desc(self):
        """
        Returns the mapping of index to description.

        Returns:
            dict[int, str]: the mapping of index to description.
        """
        return self.index2code

    @cached_property
    def code2index(self):
        """
        Returns the mapping of code to index.

        Returns:
            dict[str, int]: the mapping of code to index.
        """
        return {v: k for k, v in self.index2code.items()}

    def empty(self):
        """
        Returns an empty InpatientObservables object.

        Returns:
            InpatientObservables: an empty InpatientObservables object.
        """
        return InpatientObservables.empty(len(self.config.leading_hours))

    @staticmethod
    def _nan_concat_leading_windows(x: Array) -> Array:
        """
        Generates sliding windows of the input array, padded with NaN values.

        Args:
            x (Array): the input array.

        Returns:
            Array: the sliding windows of the input array, padded with NaN values.
        """
        n = len(x)
        add_arr = np.full(n - 1, np.nan)
        x_ext = np.concatenate((add_arr, x[::-1]))
        strided = np.lib.stride_tricks.as_strided
        nrows = len(x_ext) - n + 1
        s = x_ext.strides[0]
        return strided(x_ext, shape=(nrows, n), strides=(s, s))[::-1, ::-1]

    @staticmethod
    def filter_first_acquisitions(m: int, n: int) -> Array:
        """
        Generates a mask ignoring the first acquisitions in the input array.

        Args:
            n (int): number of acquisitions to neutralize.


        Returns:
            Array: the mask ignoring the first acquisitions in the input array.
        """
        return np.where(np.arange(m) < n, False, True).astype(bool)

    @staticmethod
    def filter_entry_neglect_window(t: Array, neglect_window: float) -> Array:
        """
        Generates a mask that ignores the observations in the beginning of the admission within `neglect_window`.
        The mask suppression is inclusive of the timestamps equal to `neglect_window`.

        Args:
            t (Array): the time array.
            neglect_window (float): number of hours to neglect in the beginning.

        Returns:
            Array: the neutralized input array.
        """
        return np.where(t <= neglect_window, False, True).astype(bool)

    @staticmethod
    def filter_recovery_window(t: Array, x: Array, recovery_window: float) -> Array:
        """
        Generates a mask the ignores the observations within the recovery window.

        Args:
            t (Array): the time array.
            x (Array): the input array.
            recovery_window (float): number of hours to neglect in the beginning.

        Returns:
            Array: the neutralized input array.
        """
        mask = np.ones_like(x).astype(bool)
        if len(t) == 0 or len(t) == 1:
            return mask
        x0 = x[0:-1]
        x1 = x[1:]
        next_recovery = (x0 != 0) & (~np.isnan(x0)) & (x1 == 0)

        for i in np.flatnonzero(next_recovery):
            mask[i + 1 :] = np.where(t[i + 1 :] - t[i] <= recovery_window, 0, 1)
        return mask

    @staticmethod
    def _nan_agg_nonzero(x: Array, axis: int | None) -> Array:
        """
        Aggregates the values in a given array along the specified axis, treating NaN values as zero.

        Args:
            x (Array): The input array.
            axis: The axis along which to aggregate.

        Returns:
            Array: The aggregated array.
        """
        all_nan = np.all(np.isnan(x), axis=axis) * 1.0
        replaced_nan = np.where(np.isnan(x), 0, x)
        return np.where(all_nan, np.nan, np.any(replaced_nan, axis=axis) * 1.0)

    @staticmethod
    def _nan_agg_max(x: Array, axis: int | None):
        """
        Aggregates the values in a given array along the specified axis, treating NaN values as zero.

        Args:
            x (Array): The input array.
            axis: The axis along which to aggregate.

        Returns:
            Array: The aggregated array.
        """
        return np.nanmax(x, axis=axis)

    @classmethod
    def aggregation(cls, aggregation: str) -> Callable:
        """
        Returns the aggregation function based on the aggregation scheme.

        Returns:
            Callable: the aggregation function.
        """
        if aggregation == "max":
            return cls._nan_agg_max
        elif aggregation == "any":
            return cls._nan_agg_nonzero
        else:
            raise ValueError(f"Aggregation {aggregation} not supported")

    @classmethod
    def mask_noisy_observations(
        cls, t: Array, x: Array, entry_neglect_window: float, recovery_window: float, minimum_acquisitions: int
    ) -> Array:
        # neutralize the first acquisitions
        m = cls.filter_first_acquisitions(len(t), minimum_acquisitions)
        # neutralize the observations in the beginning within the entry neglect window.
        m &= cls.filter_entry_neglect_window(t, entry_neglect_window)
        # neutralize the observations within the recovery window.
        m &= cls.filter_recovery_window(t, x, recovery_window)
        return m

    @classmethod
    def extract_leading_window(cls, t: Array, x: Array, leading_hours: list[float], aggregation: str) -> Array:
        agg = cls.aggregation(aggregation)
        # a time-window starting from timestamp_i for each row_i
        # if time = [t0, t1, t2, t3]
        # then t_leads = [[t0, t1, t2, t3],
        #                 [t1, t2, t3, nan],
        #                 [t2, t3, nan, nan],
        #                 [t3, nan, nan, nan]]
        t_leads = cls._nan_concat_leading_windows(t)

        # if time = [t0, t1, t2, t3]
        # then delta_leads = [[0, t1-t0, t2-t0, t3-t0],
        #                     [0, t2-t1, t3-t1, nan],
        #                     [0, t3-t2, nan, nan],
        #                     [0, nan, nan, nan]]
        delta_leads = t_leads - t_leads[:, 0, None]

        # a value-window starting from timestamp_i for each row_i
        # if value = [v0, v1, v2, v3]
        # then v_leads = [[v0, v1, v2, v3],
        #                 [v1, v2, v3, nan],
        #                 [v2, v3, nan, nan],
        #                 [v3, nan, nan, nan]]
        v_leads = cls._nan_concat_leading_windows(x)

        values = []
        for w in leading_hours:
            # select the rows where the time-window is less than t
            # if time = [t0, t1, t2, t3]
            # then mask = [[0 < t, t1 - t0 < t, t2 - t0 < t, t3 - t0 < t],
            #              [0 < t, t2 - t1 < t, t3 - t1 < t, nan < t],
            #              [0 < t, t3 - t2 < t, nan < t, nan < t]],
            #              [0 < t, nan < t, nan < t, nan < t]]
            mask = delta_leads < w
            v_lead = np.where(mask, v_leads, np.nan)
            values.append(agg(v_lead, axis=1).flatten())

        return np.stack(values, axis=1)

    def __call__(self, observables: InpatientObservables):
        """
        Makes a leading observable from the current timestamp.

        Args:
            observables (InpatientObservables): The input observables.

        Returns:
            InpatientObservables: The resulting leading observables.
        """
        if len(observables) == 0:
            return self.empty()

        assert observables.value.shape[2] == 1, "Only one channel is supported."

        time = observables.time
        value = observables.value[:, self.code_index].squeeze(axis=-1)
        mask = observables.mask[:, self.code_index]
        mask &= self.mask_noisy_observations(
            time,
            value,
            entry_neglect_window=self.config.entry_neglect_window,
            recovery_window=self.config.recovery_window,
            minimum_acquisitions=self.config.minimum_acquisitions,
        )
        value = np.where(mask, value, np.nan)
        value = self.extract_leading_window(time, value, self.config.leading_hours, self.aggregation_name)
        return InpatientObservables(time, value, mask=~np.isnan(value))


class InpatientInput(AbstractVxData):
    """
    Represents inpatient input data.

    Attributes:
        code_index (Array): the code index array. Each index value correspond to a unique input code.
        starttime (Array): the start time array.
        endtime (Array): the end time array.
        rate (Optional[Array]): the rate array, if available. If not provided, it is assumed to be a vector of ones.
    """

    code_index: Array
    starttime: Array
    endtime: Array
    rate: Array

    def __init__(self, code_index: Array, starttime: Array, endtime: Array, rate: Array | None = None):
        self.code_index = code_index
        self.starttime = starttime
        self.endtime = endtime
        self.rate = rate if rate is not None else np.ones(len(code_index), dtype=bool)

    def __check_init__(self):
        assert len(self.code_index) == len(self.starttime) == len(self.endtime)
        assert self.code_index.dtype == int and self.starttime.dtype == self.endtime.dtype
        if self.rate is not None:
            assert len(self.code_index) == len(self.rate)
        assert np.all(self.starttime <= self.endtime)
        assert np.all(self.starttime >= 0.0)

    def __len__(self):
        return len(self.code_index)

    def __call__(self, t: float, input_size: int) -> Array:
        """
        Returns the vectorized input at time t.

        Args:
            t (float): the time at which to return the input.
            input_size (int): the size of the input vector.

        Returns:
            Array: the input at time t.
        """

        mask = (self.starttime <= t) & (t < self.endtime)
        index = self.code_index[mask]
        rate = self.rate[mask]
        vec_input = np.zeros(input_size, dtype=rate.dtype)
        vec_input[index] += rate
        return vec_input

    @classmethod
    def empty(cls, code_index_dtype=int, starttime_dtype=float, endtime_dtype=float, rate_dtype=float) -> Self:
        zero_vec = np.zeros(0, dtype=bool)
        return cls(
            code_index=zero_vec.astype(code_index_dtype),
            starttime=zero_vec.astype(starttime_dtype),
            endtime=zero_vec.astype(endtime_dtype),
            rate=zero_vec.astype(rate_dtype),
        )


class InpatientInterventions(AbstractVxData):
    hosp_procedures: InpatientInput | None
    icu_procedures: InpatientInput | None
    icu_inputs: InpatientInput | None

    def __init__(
        self,
        hosp_procedures: InpatientInput | None = None,
        icu_procedures: InpatientInput | None = None,
        icu_inputs: InpatientInput | None = None,
    ):
        self.hosp_procedures = hosp_procedures
        self.icu_procedures = icu_procedures
        self.icu_inputs = icu_inputs

    @cached_property
    def timestamps(self) -> list[float]:
        timestamps = []
        for k in self.__dict__.keys():
            ii = getattr(self, k, None)
            if isinstance(ii, InpatientInput):
                timestamps.extend(ii.starttime)
                timestamps.extend(ii.endtime)
        return list(sorted(set(timestamps)))

    @cached_property
    def final_timestamp(self) -> float:
        return max(self.timestamps)


class SegmentedInpatientInterventions(AbstractVxData):
    time: Array
    icu_inputs: Array | None
    icu_procedures: Array | None
    hosp_procedures: Array | None

    def __init__(
        self,
        time: Array,
        icu_inputs: Array | None = None,
        icu_procedures: Array | None = None,
        hosp_procedures: Array | None = None,
    ):
        self.time = time
        self.icu_inputs = icu_inputs
        self.icu_procedures = icu_procedures
        self.hosp_procedures = hosp_procedures

    @classmethod
    def from_interventions(
        cls,
        inpatient_interventions: InpatientInterventions,
        terminal_time: float,
        hosp_procedures_size: int | None = None,
        icu_procedures_size: int | None = None,
        icu_inputs_size: int | None = None,
        maximum_padding: int = 100,
    ) -> Self:
        timestamps = inpatient_interventions.timestamps
        assert terminal_time >= max(timestamps, default=0.0), (
            f"Terminal time {terminal_time} should be greater than the maximum timestamp {max(timestamps, default=0.0)}"
        )
        assert min(timestamps, default=0.0) >= 0.0, (
            f"Minimum timestamp {min(timestamps, default=0.0)} should be greater than or equal to 0.0"
        )

        time = np.unique(np.array(timestamps + [0.0, terminal_time], dtype=np.float64))
        time = cls.pad_array(time, value=np.nan, maximum_padding=maximum_padding)
        t0 = time[:-1]
        hosp_procedures = None
        icu_procedures = None
        icu_inputs = None
        if inpatient_interventions.hosp_procedures is not None and hosp_procedures_size is not None:
            hosp_procedures = cls._segment(t0, inpatient_interventions.hosp_procedures, hosp_procedures_size)
        if inpatient_interventions.icu_procedures is not None and icu_procedures_size is not None:
            icu_procedures = cls._segment(t0, inpatient_interventions.icu_procedures, icu_procedures_size)
        if inpatient_interventions.icu_inputs is not None and icu_inputs_size is not None:
            icu_inputs = cls._segment(t0, inpatient_interventions.icu_inputs, icu_inputs_size)
        return cls(time=time, hosp_procedures=hosp_procedures, icu_procedures=icu_procedures, icu_inputs=icu_inputs)

    @staticmethod
    def pad_array(array: Array, maximum_padding: int = 100, value: float = 0.0) -> Array:
        """
        Pad array to be a multiple of maximum_padding. This is to
        minimize the number of jit-compiling that is made for the same functions
        when the input shape changes.

        Args:
            array (Array): the array to be padded.
            maximum_padding (int, optional): the maximum padding. Defaults to 100.
            value (float, optional): the value to pad with. Defaults to 0.0.

        Returns:
            Array: the padded array."""

        n = len(array)
        n_pad = maximum_padding - (n % maximum_padding)
        if n_pad == maximum_padding:
            return array

        return np.pad(array, pad_width=(0, n_pad), mode="constant", constant_values=value)

    @staticmethod
    def _segment(t0_padded: Array, inpatient_input: InpatientInput, input_size: int) -> Array:
        """
        Generate segmented procedures and apply the specified procedure transformation.
        Returns:
            numpy.ndarray: the processed segments.
        """
        t = t0_padded[~np.isnan(t0_padded)]
        t_nan = t0_padded[np.isnan(t0_padded)]
        out = np.stack([inpatient_input(ti, input_size) for ti in t], axis=0)
        pad = np.zeros((len(t_nan), out[0].shape[0]), dtype=out.dtype)
        return np.vstack([out, pad])

    @cached_property
    def t0_padded(self):
        return self.time[:-1]

    @property
    def t0(self):
        """Start times for segmenting the interventions"""
        t = self.time
        xnp = np_module(t)
        return t[~xnp.isnan(t)][:-1]

    @property
    def t1_padded(self):
        """End times for segmenting the interventions"""
        return self.time[1:]

    @property
    def t1(self):
        """End times for segmenting the interventions"""
        t = self.time
        xnp = np_module(t)
        return t[~xnp.isnan(t)][1:]


class AdmissionDates(AbstractVxData):
    admission: pd.Timestamp
    discharge: pd.Timestamp

    def __init__(self, admission: pd.Timestamp, discharge: pd.Timestamp):
        self.admission = admission
        self.discharge = discharge

    def __getitem__(self, item) -> pd.Timestamp:
        if item == 0:
            return self.admission
        elif item == 1:
            return self.discharge
        else:
            raise IndexError


class Admission(AbstractVxData):
    """Admission data class representing a hospital admission.

    Attributes:
        admission_id: unique ID for the admission.
        admission_dates: start and end dates for the admission.
        dx_codes: diagnosis codes recorded during the admission.
        dx_codes_history: historical diagnosis codes prior to admission.
        outcome: outcome codes of interest derived from the diagnosis codes.
        observables: timeseries clinical observations data.
        interventions: timeseries clinical interventions data.
        leading_observable: timeseries clinical leading observable (of interest).
    """

    admission_id: str  # Unique ID for each admission
    admission_dates: AdmissionDates
    dx_codes: CodesVector
    dx_codes_history: CodesVector
    outcome: CodesVector
    observables: InpatientObservables | None
    interventions: InpatientInterventions | None
    leading_observable: InpatientObservables | None

    def __init__(
        self,
        admission_id: str,
        admission_dates: AdmissionDates,
        dx_codes: CodesVector,
        dx_codes_history: CodesVector,
        outcome: CodesVector,
        observables: InpatientObservables | None = None,
        interventions: InpatientInterventions | None = None,
        leading_observable: InpatientObservables | None = None,
    ):
        self.admission_id = admission_id
        self.admission_dates = admission_dates
        self.dx_codes = dx_codes
        self.dx_codes_history = dx_codes_history
        self.outcome = outcome
        self.observables = observables
        self.interventions = interventions
        self.leading_observable = leading_observable

    def extract_leading_observable(self, leading_observable_extractor: LeadingObservableExtractor) -> Self:
        """
        Extracts the leading observable from the admission data.

        Args:
            leading_observable_extractor (LeadingObservableExtractor): the leading observable extractor.

        Returns:
            Admission: the admission data with the leading observable extracted.
        """
        assert self.observables is not None, "Observables must be provided to extract leading observable."
        leading_observable = leading_observable_extractor(self.observables)
        return type(self)(
            admission_id=self.admission_id,
            admission_dates=self.admission_dates,
            dx_codes=self.dx_codes,
            dx_codes_history=self.dx_codes_history,
            outcome=self.outcome,
            observables=self.observables,
            leading_observable=leading_observable,
            interventions=self.interventions,
        )

    def observables_time_binning(self, interval: float, types: tuple[NumericalTypeHint, ...]) -> Self:
        """
        Bins the observables data into time intervals of the specified length.

        Args:
            interval (float): the length of the time intervals.
            obs_scheme (NumericScheme): the numeric scheme for the observables, which is used to guide to aggregation
                operations.
        Returns:
            Admission: the admission data with the observables data binned.

        """
        assert self.observables is not None, "Observables must be provided to bin the time series."
        observables = self.observables.time_binning(interval, types)
        return type(self)(
            admission_id=self.admission_id,
            admission_dates=self.admission_dates,
            dx_codes=self.dx_codes,
            dx_codes_history=self.dx_codes_history,
            outcome=self.outcome,
            observables=observables,
            leading_observable=self.leading_observable,
            interventions=self.interventions,
        )

    @cached_property
    def interval_hours(self) -> float:
        """
        Calculates the interval in hours between the admission start and end dates.

        Returns:
            float: the interval in hours.
        """
        return (self.admission_dates[1] - self.admission_dates[0]).total_seconds() / 3600

    @cached_property
    def interval_days(self) -> float:
        """
        Calculates the interval in days based on the interval in hours.

        Returns:
            float: the interval in days.
        """
        return self.interval_hours / 24

    def days_since(self, date: pd.Timestamp) -> tuple[float, float]:
        """
        Calculates the number of days from a reference date to admission and discharge dates, respectively.

        Args:
            date (date): the date to calculate the number of days since.

        Returns:
            tuple[float, float]: the number of days passed at the admission and discharge dates, respectively.
        """
        d1 = (self.admission_dates[0] - date).total_seconds() / 3600 / 24
        d2 = (self.admission_dates[1] - date).total_seconds() / 3600 / 24
        return d1, d2


class SegmentedAdmission(Admission):
    observables: SegmentedInpatientObservables | None
    interventions: SegmentedInpatientInterventions | None
    leading_observable: SegmentedInpatientObservables | None

    def __init__(
        self,
        admission_id: str,
        admission_dates: AdmissionDates,
        dx_codes: CodesVector,
        dx_codes_history: CodesVector,
        outcome: CodesVector,
        observables: SegmentedInpatientObservables | None = None,
        interventions: SegmentedInpatientInterventions | None = None,
        leading_observable: SegmentedInpatientObservables | None = None,
    ):
        super().__init__(
            admission_id=admission_id,
            admission_dates=admission_dates,
            dx_codes=dx_codes,
            dx_codes_history=dx_codes_history,
            outcome=outcome,
            observables=observables,
            interventions=interventions,  # type: ignore
            leading_observable=leading_observable,
        )

    @staticmethod
    def _segment_interventions(
        interventions: InpatientInterventions | None,
        terminal_time: float,
        hosp_procedures_size: int | None,
        icu_procedures_size: int | None,
        icu_inputs_size: int | None,
        maximum_padding: int = 100,
    ) -> SegmentedInpatientInterventions | None:
        if interventions is None:
            return None
        return SegmentedInpatientInterventions.from_interventions(
            interventions,
            terminal_time,
            hosp_procedures_size=hosp_procedures_size,
            icu_procedures_size=icu_procedures_size,
            icu_inputs_size=icu_inputs_size,
            maximum_padding=maximum_padding,
        )

    @staticmethod
    def _segment_observables(
        observables: InpatientObservables | None, interventions: SegmentedInpatientInterventions | None
    ) -> SegmentedInpatientObservables | None:
        if observables is None:
            return None
        if interventions is None:
            return SegmentedInpatientObservables.from_observables(observables, np.array([]))
        time = interventions.time
        t_sep = time[~np.isnan(time)][1:-1]
        return SegmentedInpatientObservables.from_observables(observables, t_sep)

    @staticmethod
    def from_admission(
        admission: Admission,
        hosp_procedures_size: int | None,
        icu_procedures_size: int | None,
        icu_inputs_size: int | None,
        maximum_padding: int = 100,
    ) -> "SegmentedAdmission":
        interventions = SegmentedAdmission._segment_interventions(
            admission.interventions,
            admission.interval_hours,
            hosp_procedures_size=hosp_procedures_size,
            icu_procedures_size=icu_procedures_size,
            icu_inputs_size=icu_inputs_size,
            maximum_padding=maximum_padding,
        )
        observables = SegmentedAdmission._segment_observables(admission.observables, interventions)
        leading_observable = SegmentedAdmission._segment_observables(admission.leading_observable, interventions)
        return SegmentedAdmission(
            admission_id=admission.admission_id,
            admission_dates=admission.admission_dates,
            dx_codes=admission.dx_codes,
            dx_codes_history=admission.dx_codes_history,
            outcome=admission.outcome,
            observables=observables,
            interventions=interventions,
            leading_observable=leading_observable,
        )


class DemographicVectorConfig(AbstractConfig):
    """
    Configuration class for demographic vector.

    Attributes:
        gender (bool): indicates whether gender is included in the vector.
        age (bool): indicates whether age is included in the vector.
        ethnicity (bool): indicates whether ethnicity is included in the vector.
    """

    gender: bool
    age: bool
    ethnicity: bool

    def __init__(self, gender: bool = True, age: bool = True, ethnicity: bool = True):
        self.gender = gender
        self.age = age
        self.ethnicity = ethnicity


class CPRDDemographicVectorConfig(DemographicVectorConfig):
    """
    Configuration class for CPRD demographic vector.
    """

    imd: bool = True


class StaticInfo(AbstractVxData):
    """
    Represents static information about a patient.

    Attributes:
        gender (Optional[CodesVector]): gender information of the patient.
        ethnicity (Optional[CodesVector]): ethnicity information of the patient.
        date_of_birth (Optional[date]): date of birth of the patient.

    Methods:
        age: calculates the age of the patient based on the current date.
        demographic_vector: returns the demographic vector based on the current date.
        _concat: concatenates the age and vector.

    """

    gender: CodesVector | None
    ethnicity: CodesVector | None
    date_of_birth: pd.Timestamp | None

    def __init__(self, gender: CodesVector | None, ethnicity: CodesVector | None, date_of_birth: pd.Timestamp | None):
        self.gender = gender
        self.ethnicity = ethnicity
        self.date_of_birth = date_of_birth

    def constant_attrs_list(self, demographic_vector_config: DemographicVectorConfig) -> list[Array]:
        attrs_vec = []
        if demographic_vector_config.gender:
            assert self.gender is not None and len(self.gender.scheme) > 0, "Gender is not extracted from the dataset"
            attrs_vec.append(self.gender.vec)
        if demographic_vector_config.ethnicity:
            assert self.ethnicity is not None, "Ethnicity is not extracted from the dataset"
            attrs_vec.append(self.ethnicity.vec)
        return attrs_vec

    @functools.lru_cache(maxsize=10, typed=True)
    def constant_vec(self, demographic_vector_config: DemographicVectorConfig) -> Array:
        attrs_vec = self.constant_attrs_list(demographic_vector_config)
        xnp = np_module(attrs_vec[0])
        if len(attrs_vec) == 0:
            return xnp.array([], dtype=xnp.float16)
        else:
            return xnp.hstack(attrs_vec)

    def age(self, current_date: pd.Timestamp) -> float:
        """
        Calculates the age of the patient based on the current date.

        Args:
            current_date (date): the current date.

        Returns:
            float: the age of the patient in years.
        """
        assert self.date_of_birth is not None, "Date of birth is not set"
        return current_date.year - self.date_of_birth.year

    def admission_demographics(self, admission: Admission, demographic_vector_config: DemographicVectorConfig) -> Array:
        """
        Returns the demographic vector for the patient.

        Args:
            demographic_vector_config (DemographicVectorConfig): the demographic vector configuration.

        Returns:
            Array: the demographic vector.
        """
        if demographic_vector_config.age:
            return jnp.hstack((self.age(admission.admission_dates[0]), self.constant_vec(demographic_vector_config)))
        return self.constant_vec(demographic_vector_config)


class Patient(AbstractVxData):
    """
    Represents a patient with demographic information and a clinical history
    of admissions.

    Attributes:
        subject_id: Unique identifier for the patient.
        static_info: Static demographic and geographic data for the patient.
        admissions: list of hospital/clinic admissions for the patient.


    Properties:
        d2d_interval_days: number of days between first and last discharge.


    Methods:
        outcome_frequency_vec: accumulates outcome vectors from all admissions to get
        aggregate vector for patient.
    """

    subject_id: str
    static_info: StaticInfo
    admissions: list[Admission]

    admission_cls: ClassVar[type[Admission]] = Admission

    def __init__(self, subject_id: str, static_info: StaticInfo, admissions: list[Admission]):
        self.subject_id = subject_id
        self.static_info = static_info
        self.admissions = list(sorted(admissions, key=lambda x: x.admission_dates[0]))

    def __len__(self):
        return len(self.admissions)

    def admission_demographics(self, demographic_vector_config: DemographicVectorConfig) -> dict[str, Array]:
        """
        Returns the demographic vector for the patient.

        Args:
            demographic_vector_config (DemographicVectorConfig): the demographic vector configuration.

        Returns:
            Array: the demographic vector.
        """
        return {
            a.admission_id: self.static_info.admission_demographics(a, demographic_vector_config)
            for a in self.admissions
        }

    @cached_property
    def d2d_interval_days(self):
        """
        The interval in days between the first and last discharge dates.

        Returns:
            float: the interval in days.
        """
        d1 = self.admissions[0].admission_dates[1]
        d2 = self.admissions[-1].admission_dates[1]
        return (d2 - d1).total_seconds() / 3600 / 24

    def filter_short_stays(self, min_hours: float) -> Self:
        """
        Filter out admissions that are shorter than a specified duration.

        Args:
            min_hours (float): the minimum duration in hours.
        """
        admissions = [a for a in self.admissions if a.interval_hours >= min_hours]
        return type(self)(subject_id=self.subject_id, static_info=self.static_info, admissions=admissions)

    def outcome_frequency_vec(self) -> Array:
        """
        Calculate the accumulation of outcome vectors for all admissions.

        Returns:
            The total count of outcome code occurrence over all admissions.
        """
        return sum(a.outcome.vec for a in self.admissions)

    def __eq__(self, other) -> bool:
        return self.equals(other)

    def extract_leading_observables(self, leading_observable_extractor: LeadingObservableExtractor) -> Self:
        """
        Extracts the leading observable from all admissions.

        Args:
            leading_observable_extractor (LeadingObservableExtractor): the leading observable extractor.

        Returns:
            Patient: the patient data with the leading observable extracted.
        """
        admissions = [a.extract_leading_observable(leading_observable_extractor) for a in self.admissions]
        return type(self)(subject_id=self.subject_id, static_info=self.static_info, admissions=admissions)

    def observables_time_binning(self, interval: float, types: tuple[NumericalTypeHint, ...]) -> Self:
        """
        Bins the observables for all admissions.

        Args:
            interval (float): the interval in hours to bin the observables.

        Returns:
            Patient: the patient data with the observables binned.
        """
        admissions = [a.observables_time_binning(interval, types) for a in self.admissions]
        return type(self)(subject_id=self.subject_id, static_info=self.static_info, admissions=admissions)


class SegmentedPatient(Patient):
    admissions: list[SegmentedAdmission]
    admission_cls: ClassVar[type[SegmentedAdmission]] = SegmentedAdmission

    def __init__(self, subject_id: str, static_info: StaticInfo, admissions: list[SegmentedAdmission]):
        super().__init__(
            subject_id=subject_id,
            static_info=static_info,
            admissions=admissions,  # type: ignore
        )

    def extract_leading_observables(self, leading_observable_extractor: LeadingObservableExtractor) -> Patient:
        raise NotImplementedError("SegmentedPatient does not support leading observable extraction")

    @staticmethod
    def from_patient(
        patient: Patient,
        hosp_procedures_size: int | None,
        icu_procedures_size: int | None,
        icu_inputs_size: int | None,
        maximum_padding: int = 100,
    ) -> "SegmentedPatient":
        admissions = [
            SegmentedAdmission.from_admission(
                a,
                hosp_procedures_size=hosp_procedures_size,
                icu_procedures_size=icu_procedures_size,
                icu_inputs_size=icu_inputs_size,
                maximum_padding=maximum_padding,
            )
            for a in patient.admissions
        ]
        return SegmentedPatient(subject_id=patient.subject_id, static_info=patient.static_info, admissions=admissions)



from abc import abstractmethod
from collections.abc import Callable
from typing import get_args
from unittest import mock

import ehrax as rx
import equinox as eqx
import numpy as np
import pandas as pd
import pytest
import tables as tb
from ehrax._literals import TableAggregationLiteral
from ehrax._stats.dataset import DatasetStatsInterface, TargetHistogram
from ehrax.dataset import AdmissionsTableColumns
from ehrax.testing.common_setup import DATASET_SCHEME_CONF, DATASET_SCHEME_MANAGER, OUTCOME_DATA


@pytest.mark.parametrize(
    "columns, id_cols, code_cols, time_cols, index",
    [
        [rx.dataset.StaticTableColumns(), ("subject_id",), (), ("date_of_birth",), ("subject_id",)],
        [
            rx.dataset.AdmissionsTableColumns(),
            ("subject_id", "admission_id"),
            (),
            ("start_time", "end_time"),
            ("admission_id",),
        ],
        [rx.dataset.AdmissionSummaryTableColumns(), ("admission_id",), ("code",), (), ()],
        [rx.dataset.AdmissionTimeSeriesTableColumns(), ("admission_id",), ("code",), ("time",), ()],
        [
            rx.dataset.AdmissionIntervalEventsTableColumns(),
            ("admission_id",),
            ("code",),
            ("start_time", "end_time"),
            (),
        ],
        [rx.dataset.AdmissionIntervalRatesTableColumns(), ("admission_id",), ("code",), ("start_time", "end_time"), ()],
    ],
)
def test_table_config(
    columns: rx.dataset.TableColumns,
    id_cols: tuple[str, ...],
    code_cols: tuple[str, ...],
    time_cols: tuple[str, ...],
    index: str | None,
):
    assert sorted(columns.time_cols) == sorted(time_cols)
    assert sorted(columns.code_cols) == sorted(code_cols)
    assert sorted(columns.id_cols) == sorted(id_cols)
    assert sorted(columns.index) == sorted(index)


def test_assert_invalid_column_fail():
    with pytest.raises(AssertionError, match="Fields must be one of"):
        c = rx.dataset.DatasetColumns()
        updated = eqx.tree_at(lambda x: x.admissions.subject_id, c, "subject_id_")
        updated.validate()


def test_scheme_dict():
    scheme = rx.DatasetSchemeProxy(config=DATASET_SCHEME_CONF, schemes_context=DATASET_SCHEME_MANAGER)

    for space, scheme_name in DATASET_SCHEME_CONF.as_dict().items():
        assert hasattr(scheme, space)
        if scheme_name is not None:
            assert isinstance(getattr(scheme, space), rx.CodingScheme)


class AbstractTestDataset:
    @pytest.fixture(scope="class")
    @abstractmethod
    def dataset_tables(self, *args) -> rx.DatasetTables:
        raise NotImplementedError()

    @pytest.fixture(scope="class")
    @abstractmethod
    def dataset(self, *args) -> rx.Dataset:
        raise NotImplementedError()

    @abstractmethod
    def pipeline(self) -> list[rx.DatasetTransformation]:
        raise NotImplementedError()

    @pytest.fixture(scope="class")
    def processed_dataset(self, dataset: rx.Dataset, pipeline: list[rx.DatasetTransformation]) -> rx.Dataset:
        return dataset._execute_pipeline(pipeline, DATASET_SCHEME_MANAGER)

    @pytest.fixture(scope="class")
    def dataset_with_zero_pipeline(self, dataset: rx.Dataset):
        return dataset.execute_pipeline(rx.AbstractDatasetPipeline(config=None, transformations=[]), None)

    def test_tables_dict_property(self, dataset_tables: rx.DatasetTables):
        all_tables_keys = (
            "static",
            "admissions",
            "dx_discharge",
            "obs",
            "icu_procedures",
            "icu_inputs",
            "hosp_procedures",
        )

        assert set(dataset_tables.tables_dict.keys()) == set(
            k for k in all_tables_keys if getattr(dataset_tables, k) is not None
        )

    def test_save_load_tables(self, dataset_tables: rx.DatasetTables, tmpdir):
        with tb.open_file(f"{tmpdir}/test_dataset_tables.h5", "w") as hf5:
            dataset_tables.save(hf5.create_group("/", "dataset_tables"))
        with tb.open_file(f"{tmpdir}/test_dataset_tables.h5", "r") as hf5:
            loaded = rx.DatasetTables.load(hf5.root["dataset_tables"])
        assert loaded.equals(dataset_tables)

    def test_execute_pipeline(self, dataset: rx.Dataset, dataset_with_zero_pipeline: rx.Dataset):
        assert isinstance(dataset, rx.Dataset)
        assert isinstance(dataset_with_zero_pipeline, rx.Dataset)
        assert dataset.pipeline_report.equals(pd.DataFrame())

        # Because we use identity pipeline, the dataset columns should be the same
        # but the new dataset should have a different report (metadata).
        assert not dataset_with_zero_pipeline.equals(dataset)
        assert not dataset_with_zero_pipeline.pipeline_report.equals(dataset.pipeline_report)
        assert dataset_with_zero_pipeline.tables.equals(dataset.tables)
        assert len(dataset_with_zero_pipeline.pipeline_report) == 1
        assert dataset_with_zero_pipeline.pipeline_report.loc[0, "transformation"] == "identity"

        with mock.patch("logging.warning") as mocker:
            dataset3 = dataset_with_zero_pipeline.execute_pipeline(
                rx.AbstractDatasetPipeline(config=None, transformations=[]), None
            )
            assert dataset3.equals(dataset_with_zero_pipeline)
            mocker.assert_called_once_with("A pipeline has already been executed. Doing nothing.")

    def test_subject_ids_of_unindexed_dataset(self, dataset: rx.Dataset, dataset_with_zero_pipeline: rx.Dataset):
        with pytest.raises(AssertionError):
            _ = dataset.subject_ids

        with pytest.raises(AssertionError):
            _ = dataset_with_zero_pipeline.subject_ids

    def test_subject_ids_of_indexed_dataset(self, processed_dataset: rx.Dataset):
        assert set(processed_dataset.subject_ids) == set(processed_dataset.tables.static.index.unique())

    @pytest.mark.expensive_test
    def test_save_load(self, dataset: rx.Dataset, dataset_with_zero_pipeline: rx.Dataset, tmpdir: str):
        dataset_with_zero_pipeline.save(f"{tmpdir}/test_dataset")
        loaded = rx.Dataset.load(f"{tmpdir}/test_dataset")
        assert loaded.equals(dataset_with_zero_pipeline)
        assert not loaded.equals(dataset)
        assert loaded.equals(dataset._execute_pipeline([], None))

    @pytest.fixture(scope="class")
    def subject_ids(self, processed_dataset: rx.Dataset):
        return processed_dataset.subject_ids

    @pytest.mark.parametrize("valid_split", [[1.0]])
    @pytest.mark.parametrize("valid_balance", ["subjects", "admissions", "admissions_intervals"])
    @pytest.mark.parametrize("invalid_splits", [[], [0.3, 0.8, 0.7, 0.9], [0.5, 0.2]])  # should be sorted.
    @pytest.mark.parametrize("invalid_balance", ["hi", "unsupported"])
    def test_random_split_invalid_args(
        self,
        processed_dataset: rx.Dataset,
        subject_ids: list[str],
        valid_split: list[float],
        valid_balance: rx.SplitLiteral,
        invalid_splits: list[float],
        invalid_balance: rx.SplitLiteral,
    ):
        if len(subject_ids) == 0:
            with pytest.raises(AssertionError):
                processed_dataset.random_splits(valid_split, balance=valid_balance)
            return

        if len(processed_dataset.tables.admissions) == 0 and "admissions" in valid_balance:
            with pytest.raises(AssertionError):
                processed_dataset.random_splits(valid_split, balance=valid_balance)
            return

        assert set(processed_dataset.random_splits(valid_split, balance=valid_balance)[0]) == set(subject_ids)

        with pytest.raises(AssertionError):
            processed_dataset.random_splits(valid_split, balance=invalid_balance)

        with pytest.raises(AssertionError):
            processed_dataset.random_splits(invalid_splits, balance=valid_balance)

        with pytest.raises(AssertionError):
            processed_dataset.random_splits(invalid_splits, balance=invalid_balance)


class TestDatasetWithoutRecords(AbstractTestDataset):
    @pytest.fixture(scope="class")
    def dataset_tables(self, dataset_tables_without_records: rx.DatasetTables) -> rx.DatasetTables:
        return dataset_tables_without_records

    @pytest.fixture(scope="class")
    def dataset(self, dataset_without_records: rx.Dataset) -> rx.Dataset:
        return dataset_without_records

    @pytest.fixture(scope="class")
    def pipeline(self) -> list[rx.DatasetTransformation]:
        return [rx.SetIndex(), rx.SynchronizeSubjects(), rx.CastTimestamps(), rx.SetAdmissionRelativeTimes()]


class TestDatasetWithRecords(AbstractTestDataset):
    @pytest.fixture(scope="class")
    def dataset_tables(self, dataset_tables_with_records: rx.DatasetTables) -> rx.DatasetTables:
        return dataset_tables_with_records

    @pytest.fixture(scope="class")
    def dataset(self, dataset_with_records: rx.Dataset) -> rx.Dataset:
        return dataset_with_records

    @pytest.fixture(scope="class")
    def pipeline(self):
        return [
            rx.SetIndex(),
            rx.SynchronizeSubjects(),
            rx.CastTimestamps(),
            rx.ICUInputRateUnitConversion(),
            rx.SetAdmissionRelativeTimes(),
        ]

    @pytest.fixture(scope="class")
    def split_measure(self, processed_dataset: rx.Dataset, balance: str):
        return {
            "subjects": lambda x: len(x),
            "admissions": lambda x: sum(processed_dataset.subjects_n_admissions.loc[x]),
            "admissions_intervals": lambda x: sum(processed_dataset.subjects_intervals_sum.loc[x]),
        }[balance]

    @pytest.fixture(params=["subjects", "admissions", "admissions_intervals"], scope="class")
    def balance(self, request):
        return request.param

    @pytest.fixture(params=[[0.1], [0.1, 0.5, 0.7, 0.9]], scope="class")
    def split_quantiles(self, request):
        return request.param

    @pytest.fixture(params=[1, 11, 111], scope="class")
    def subject_splits(
        self, processed_dataset: rx.Dataset, balance: rx.SplitLiteral, split_quantiles: list[float], request
    ):
        random_seed = request.param
        return processed_dataset.random_splits(split_quantiles, balance=balance, random_seed=random_seed)

    def test_random_split(
        self,
        processed_dataset: rx.Dataset,
        subject_ids: list[str],
        subject_splits: list[list[str]],
        split_quantiles: list[float],
    ):
        assert set(i for ii in subject_splits for i in ii) == set(subject_ids)
        assert len(subject_splits) == len(split_quantiles) + 1
        # No overlaps.
        assert sum(len(v) for v in subject_splits) == len(processed_dataset.subject_ids)
        assert set(i for ii in subject_splits for i in ii) == set(processed_dataset.subject_ids)

    @pytest.fixture
    def split_proportions(self, split_quantiles: list[float]):
        return [p1 - p0 for p0, p1 in zip([0] + split_quantiles, split_quantiles + [1])]

    def test_random_split_balance(
        self,
        subject_ids: list[str],
        subject_splits: list[list[str]],
        split_proportions: list[float],
        balance: str,
        split_measure: Callable[[list[str]], float],
    ):
        if len(subject_ids) < 5:
            raise pytest.skip("Not enough subjects to test random split")

        # # test proportionality
        # NOTE: no specified behaviour when splits have equal proportions, so comparing argsorts
        # is not appropriate.
        p_threshold = 1 / len(subject_ids)
        tolerance = max(
            abs(split_measure([i]) - split_measure([j])) for (i, j) in zip(subject_ids[1:], subject_ids[:-1])
        )
        for i in range(len(split_proportions)):
            split_measure_i = split_measure(subject_splits[i])
            for j in range(i + 1, len(split_proportions)):
                if abs(split_proportions[i] - split_proportions[j]) < 2 * p_threshold:
                    if balance == "subjects":
                        # Difference between subjects is at most 1 when balance is applied
                        # on subjects count AND proportions are (almost) equal.
                        assert abs(len(subject_splits[i]) - len(subject_splits[j])) <= 1
                elif split_proportions[i] > split_proportions[j]:
                    assert (split_measure_i - split_measure(subject_splits[j])) >= -tolerance
                else:
                    assert (split_measure_i - split_measure(subject_splits[j])) <= tolerance

    #
    # @pytest.fixture
    # def dataset_hist_interface(self, processed_dataset: rx.Dataset) -> TargetHistogram:
    #     return processed_dataset.stats(DATASET_SCHEME_MANAGER).target_hist
    #
    # #
    # def stats_dx_discharge(self) -> :


class TestTargetHistogram:
    @pytest.fixture(scope="class")
    def stats_interface(self, dataset_with_records: rx.Dataset) -> DatasetStatsInterface:
        return dataset_with_records._execute_pipeline(
            [rx.SetIndex(), rx.SynchronizeSubjects(), rx.CastTimestamps()], DATASET_SCHEME_MANAGER
        ).stats(DATASET_SCHEME_MANAGER)

    @pytest.fixture(scope="class")
    def code_map(self) -> rx.CodeMap:
        m = rx.CodeMap(
            source_name="x",
            target_name="y",
            data=rx.FrozenDict1N({"a": {"x"}, "b": {"x"}, "c": {"y"}, "d": {"x", "y"}}),
        )
        assert m.domain == {"a", "b", "c", "d"}
        assert m.range == {"x", "y"}
        return m

    @pytest.fixture(scope="class")
    def c_code(self) -> str:
        return "code"

    @pytest.fixture(scope="class")
    def c_admission_id(self) -> str:
        return "admission_id"

    @pytest.mark.parametrize(
        "admission_id, code, expected",
        [
            # empty table -> {}
            [(), (), {}],
            # table with no valid code -> {}
            [("1",), ("!",), {}],
            [("1", "2"), ("?", "!"), {}],
            [("1", "1"), ("$", "*"), {}],
            # table with just one valid code.
            [("1",), ("a",), {"x": 1}],
            # aggregation of admissions
            [("1", "1"), ("a", "b"), {"x": 1}],  # a and b both map to x
            [("1", "1"), ("a", "b"), {"x": 1}],
            [("1", "2"), ("a", "a"), {"x": 2}],
            [("1", "2"), ("a", "b"), {"x": 2}],
            [("1", "1"), ("a", "c"), {"x": 1, "y": 1}],
            # rest
            [("1", "2", "3"), ("a", "a", "a"), {"x": 3}],
            [("1", "1", "2"), ("d", "a", "d"), {"x": 2, "y": 2}],
        ],
    )
    def test_hardcoded_table(
        self, admission_id: tuple[str, ...], code: tuple[str, ...], expected: dict[str, int], code_map: rx.CodeMap
    ):
        table = pd.DataFrame({"id": admission_id, "code": code})
        hist = TargetHistogram.compute(table, c_admission_id="id", c_code="code", scheme_mapper=code_map).to_dict()
        assert hist == expected

    @pytest.mark.parametrize(
        "admission_id, subject_id, admission_time, expected_agg_admission, "
        "expected_agg_first_admission, expected_agg_subject",
        [
            [(), (), (), (), (), ()],
            [("1",), ("a",), (0,), ("1",), ("1",), ("a",)],
            [("1", "2"), ("a", "a"), (1, 0), ("1", "2"), ("2",), ("a", "a")],
            [
                ("1", "2", "3", "4"),
                ("a", "a", "b", "b"),
                (0, 1, 4, 3),
                ("1", "2", "3", "4"),
                ("1", "4"),
                ("a", "a", "b", "b"),
            ],
        ],
    )
    def test_aggregation_adapt(
        self,
        admission_id: tuple[str, ...],
        subject_id: tuple[str, ...],
        admission_time: tuple[int, ...],
        expected_agg_admission: tuple[str, ...],
        expected_agg_first_admission: tuple[str, ...],
        expected_agg_subject: tuple[str, ...],
    ):
        adm_cols = AdmissionsTableColumns()
        table = pd.DataFrame(
            {adm_cols.admission_id: admission_id, adm_cols.subject_id: subject_id, adm_cols.start_time: admission_time}
        )
        adms = table.set_index(adm_cols.admission_id)
        adapted1, n1 = TargetHistogram.adapt_aggregation_level(
            adms, adm_cols, table, adm_cols.admission_id, "admission"
        )
        adapted2, n2 = TargetHistogram.adapt_aggregation_level(
            adms, adm_cols, table, adm_cols.admission_id, "first_admission"
        )
        adapted3, n3 = TargetHistogram.adapt_aggregation_level(adms, adm_cols, table, adm_cols.admission_id, "subject")
        assert adapted1[adm_cols.admission_id].tolist() == list(expected_agg_admission)
        assert adapted2[adm_cols.admission_id].tolist() == list(expected_agg_first_admission)
        assert adapted3[adm_cols.admission_id].tolist() == list(expected_agg_subject)
        assert n1 == len(admission_id)
        assert n2 == n3 == len(set(subject_id))

    def test_dataset_stats(self, stats_interface: DatasetStatsInterface):
        assert isinstance(stats_interface, DatasetStatsInterface)
        o = DATASET_SCHEME_MANAGER.outcome_data[OUTCOME_DATA.name]
        o_scheme = o.as_coding_scheme(DATASET_SCHEME_MANAGER.scheme[o.base_name])
        dx_scheme = DATASET_SCHEME_MANAGER.scheme[o.base_name]
        dx_stats, dx_norm = zip(
            *[stats_interface.target_hist.dx_discharge(dx_scheme.name, a) for a in get_args(TableAggregationLiteral)]
        )
        o_stats, o_norm = zip(
            *[stats_interface.target_hist.outcome(o.name, a) for a in get_args(TableAggregationLiteral)]
        )
        assert all(isinstance(dx_stats_i, pd.Series) for dx_stats_i in dx_stats)
        assert all(dx_stats_i.index.tolist() == list(dx_scheme.codes) for dx_stats_i in dx_stats)
        assert all(dx_stats_i.dtype is np.dtype("int") for dx_stats_i in dx_stats)

        assert all(isinstance(o_stats_i, pd.Series) for o_stats_i in o_stats)
        assert all(o_stats_i.index.tolist() == list(o_scheme.codes) for o_stats_i in o_stats)
        assert all(o_stats_i.dtype is np.dtype("int") for o_stats_i in o_stats)



import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict

import equinox as eqx
import numpy as np
import pandas as pd

from .coding_scheme import CodingSchemesManager, CodingSchemeWithUOM
from .dataset import (
    AbstractTransformation,
    COLUMN,
    Dataset,
    Report,
    SECONDS_TO_DAYS_SCALER,
    SECONDS_TO_HOURS_SCALER,
)


class DatasetTransformation(AbstractTransformation, metaclass=ABCMeta):
    @staticmethod
    def synchronize_index(
        dataset: Dataset, indexed_table_name: str, index_name: str, report: Report
    ) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict

        target_tables = {  # columns that have admission_id as column
            k: v for k, v in tables_dict.items() if k != indexed_table_name and index_name in v.columns
        }

        index = tables_dict[indexed_table_name].index
        tables = dataset.tables
        for table_name, table in target_tables.items():
            n1 = len(table)
            table = table[table[index_name].isin(pd.Series(index))]
            n2 = len(table)
            report = report.add(
                table=table_name, column=index_name, before=n1, after=n2, value_type="count", operation="sync_index"
            )
            tables = eqx.tree_at(lambda x: getattr(x, table_name), tables, table)

        return eqx.tree_at(lambda x: x.tables, dataset, tables), report

    @staticmethod
    def filter_subjects_with_less_than_n_admissions(
        static: pd.DataFrame, admissions: pd.DataFrame, n: int
    ) -> pd.DataFrame:
        n_admissions = static.index.map(admissions.groupby(COLUMN.subject_id).size()).fillna(0)
        return static.loc[n_admissions >= n]

    @staticmethod
    def filter_no_admission_subjects(dataset: Dataset, report: Report) -> tuple[Dataset, Report]:
        static = dataset.tables.static
        admissions = dataset.tables.admissions
        n1 = len(static)
        static = DatasetTransformation.filter_subjects_with_less_than_n_admissions(static, admissions, 1)
        n2 = len(static)
        report = report.add(
            table="static",
            column=static.index.name,
            before=n1,
            after=n2,
            value_type="count",
            operation="filter_no_admission_subjects",
        )
        return eqx.tree_at(lambda x: x.tables.static, dataset, static), report

    @classmethod
    def synchronize_admissions(cls, dataset: Dataset, report: Report) -> tuple[Dataset, Report]:
        dataset, report = cls.synchronize_index(
            dataset, "admissions", dataset.config.columns.admissions.admission_id, report
        )
        return cls.filter_no_admission_subjects(dataset, report)

    @classmethod
    def synchronize_subjects(cls, dataset: Dataset, report: Report) -> tuple[Dataset, Report]:
        # Synchronizing subjects might entail synchronizing admissions, so we need to call it first
        dataset, report = cls.synchronize_index(dataset, "static", dataset.config.columns.static.subject_id, report)
        return cls.synchronize_admissions(dataset, report)

    @classmethod
    @abstractmethod
    def apply(
        cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report
    ) -> tuple[Dataset, Report]: ...


class SynchronizeAdmissions(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        return cls.synchronize_admissions(dataset, report)


class SynchronizeSubjects(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        return cls.synchronize_subjects(dataset, report)


class SetIndex(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        for indexed_table_name, index_name, table in (
            (table_name, index_name, tables_dict[table_name])
            for table_name, index_name in dataset.config.columns.indices.items()
            if table_name in tables_dict
        ):
            (index_name,) = index_name
            index1 = table.index.name
            table = table.set_index(index_name)
            index2 = table.index.name
            report = report.add(
                table=indexed_table_name,
                column=index_name,
                before=index1,
                after=index2,
                value_type="index_name",
                operation="set_index",
            )
            dataset = eqx.tree_at(lambda x: getattr(x.tables, indexed_table_name), dataset, table)
        return dataset, report


class CastTimestamps(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        tables = dataset.tables
        tables_dict = tables.tables_dict
        for table_name, time_cols, table in (
            (name, cols, tables_dict[name])
            for name, cols in dataset.config.columns.time_cols.items()
            if name in tables_dict
        ):
            table = table.iloc[:, :]
            for time_col in time_cols:
                assert time_col in table.columns, f"{time_col} not found in {table_name}"
                assert isinstance(time_col, str)

                if table[time_col].dtype == "datetime64[ns]":
                    logging.debug(f"{table_name}[{time_col}] already in datetime64[ns]")
                    continue
                dtype1 = table[time_col].dtype
                table[time_col] = pd.to_datetime(table[time_col], errors="raise")
                dtype2 = table[time_col].dtype
                report = report.add(
                    table=table_name, column=time_col, before=dtype1, after=dtype2, value_type="dtype", operation="cast"
                )

            tables = eqx.tree_at(lambda x: getattr(x, table_name), tables, table)
        return eqx.tree_at(lambda x: x.tables, dataset, tables), report


class SqueezeToStandardColumns(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        for table_name, columns, table in (
            (name, cols, tables_dict[name])
            for name, cols in dataset.config.columns.columns_dict().items()
            if name in tables_dict
        ):
            columns = tuple(c for c in columns if c in table.columns)
            if columns == tuple(table.columns):
                continue
            elif set(columns) == set(table.columns):
                operation = "columns reordering"
            else:
                operation = "columns subsetting & reordering"

            report = report.add(
                table=table_name,
                before=", ".join(table.columns),
                after=", ".join(columns),
                value_type="columns",
                operation=operation,
            )
            dataset = eqx.tree_at(lambda x: getattr(x.tables, table_name), dataset, table[list(columns)])
        return dataset, report


class SetAdmissionRelativeTimes(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        time_cols = {
            k: v
            for k, v in dataset.config.columns.time_cols.items()
            if dataset.config.columns.temporal_admission_linked_table(k)
        }

        c_admittime = dataset.config.columns.admissions.start_time
        c_admission_id = dataset.config.columns.admissions.admission_id
        admissions = dataset.tables.admissions[[c_admittime]]
        tables_dict = dataset.tables.tables_dict

        for table_name, table_time_cols, table in (
            (name, cols, tables_dict[name]) for name, cols in time_cols.items() if name in tables_dict
        ):
            df = pd.merge(
                table, admissions, left_on=c_admission_id, right_index=True, suffixes=(None, "_y"), how="left"
            )
            admittime_col = f"{c_admittime}_y" if c_admittime in table.columns else c_admittime
            for time_col in table_time_cols:
                update = {time_col: (df[time_col] - df[admittime_col]).dt.total_seconds() * SECONDS_TO_HOURS_SCALER}
                df = df.assign(**update)
                report = report.add(
                    table=table_name,
                    column=time_col,
                    before=table[time_col].dtype,
                    after=df[time_col].dtype,
                    value_type="dtype",
                    operation="set_admission_relative_times",
                )

            df = df[table.columns]
            dataset = eqx.tree_at(lambda x: getattr(x.tables, table_name), dataset, df)

        return dataset, report


class FilterSubjectsNegativeAdmissionLengths(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        table_config = dataset.config.columns.admissions
        c_subject_id = table_config.subject_id
        c_dischtime = table_config.end_time
        c_admittime = table_config.start_time
        adms = dataset.tables.admissions

        # assert dtypes are datetime64[ns]
        assert adms[c_admittime].dtype == "datetime64[ns]" and adms[c_dischtime].dtype == "datetime64[ns]", (
            f"{c_admittime} and {c_dischtime} must be datetime64[ns]"
        )

        static = dataset.tables.static
        neg_los_sub = adms.loc[adms.loc[:, c_dischtime] < adms.loc[:, c_admittime], c_subject_id].unique()
        n_before = len(static)
        static = static[~static.index.isin(neg_los_sub)]
        n_after = len(static)
        report = report.add(
            table="static", column=c_subject_id, value_type="count", operation="filter", before=n_before, after=n_after
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class FilterUnsupportedCodes(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        for table_name, code_column, table in (
            (name, col, tables_dict[name])
            for name, col in dataset.config.columns.code_column.items()
            if name in tables_dict
        ):
            (code_column,) = code_column
            coding_scheme = getattr(dataset.scheme_proxy(schemes_context), table_name)
            n1 = len(table)
            table = table[table[code_column].isin(coding_scheme.codes)]
            n2 = len(table)
            report = report.add(
                table=table_name, column=code_column, before=n1, after=n2, value_type="count", operation="filter"
            )
            dataset = eqx.tree_at(lambda x: getattr(x.tables, table_name), dataset, table)
        return dataset, report


class FilterAdmissionsWithNoDiagnoses(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        dx_discharge = dataset.tables.dx_discharge
        assert isinstance(dx_discharge, pd.DataFrame)
        admissions = dataset.tables.admissions
        c_admission_id = dataset.config.columns.dx_discharge.admission_id
        selected_admission_id = set(dx_discharge[c_admission_id].tolist())
        n1 = len(admissions)
        admissions = admissions[admissions.index.isin(selected_admission_id)]
        n2 = len(admissions)
        report = report.add(
            table="admissions",
            column=admissions.index.name,
            before=n1,
            after=n2,
            value_type="count",
            operation="filter",
        )
        dataset = eqx.tree_at(lambda x: x.tables.admissions, dataset, admissions)
        return cls.synchronize_admissions(dataset, report)


class FilterAdmissionsWithNoObservables(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        obs = dataset.tables.obs
        assert isinstance(obs, pd.DataFrame)
        admissions = dataset.tables.admissions
        c_admission_id = dataset.config.columns.obs.admission_id
        selected_admission_id = set(obs[c_admission_id].tolist())
        n1 = len(admissions)
        admissions = admissions[admissions.index.isin(selected_admission_id)]
        n2 = len(admissions)
        report = report.add(
            table="admissions",
            column=admissions.index.name,
            before=n1,
            after=n2,
            value_type="count",
            operation="filter",
        )
        dataset = eqx.tree_at(lambda x: x.tables.admissions, dataset, admissions)
        return cls.synchronize_admissions(dataset, report)


class FilterSubjectsWithSingleOrNoAdmission(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        n1 = len(dataset.tables.static)
        static = cls.filter_subjects_with_less_than_n_admissions(dataset.tables.static, dataset.tables.admissions, 2)
        n2 = len(static)
        report = report.add(
            table="static", column=static.index.name, before=n1, after=n2, value_type="count", operation="filter"
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class FilterSubjectsWithLongAdmission(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        max_days = dataset.config.select_subjects_with_short_admissions
        if max_days is None:
            return cls.skip(dataset, report, reason="select_subjects_with_short_admissions is not configured.")

        a_df = dataset.tables.admissions.copy()
        static = dataset.tables.static
        a_df["los"] = (a_df[COLUMN.end_time] - a_df[COLUMN.start_time]).dt.total_seconds() * SECONDS_TO_DAYS_SCALER
        max_admission_los = a_df.groupby(COLUMN.subject_id)["los"].max().loc[static.index]
        n1 = len(static)
        static = static[max_admission_los < max_days]
        n2 = len(static)
        report = report.add(
            table="static", column=static.index.name, before=n1, after=n2, value_type="count", operation="filter"
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class FilterShortAdmissions(DatasetTransformation):  # without removing the corresponding subjects.
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        min_days = dataset.config.admission_minimum_los
        if min_days is None:
            return cls.skip(dataset, report, reason="select_subjects_with_short_admissions is not configured.")
        a_df = dataset.tables.admissions
        los = (a_df[COLUMN.end_time] - a_df[COLUMN.start_time]).dt.total_seconds() * SECONDS_TO_DAYS_SCALER
        n1 = len(a_df)
        a_df = a_df.loc[los >= min_days]
        n2 = len(a_df)
        report = report.add(
            table="admissions", column=a_df.index.name, before=n1, after=n2, value_type="count", operation="filter"
        )
        dataset = eqx.tree_at(lambda x: x.tables.admissions, dataset, a_df)
        return cls.synchronize_admissions(dataset, report)


class ProcessOverlappingAdmissions(DatasetTransformation):
    @staticmethod
    def _collect_overlaps(admissions: pd.DataFrame) -> dict[str, str]:
        """
        Collect overlapping admissions for a subject.
        """
        # Sort by admission time.
        if len(admissions) == 0:
            return dict()

        admissions = admissions.sort_values(COLUMN.start_time)
        intervals = list(zip(admissions.index, admissions[COLUMN.start_time], admissions[COLUMN.end_time]))
        new_admission = dict()
        new_admission_id, _, last_discharge = intervals[0]
        for admission_id, admission_time, discharge_time in intervals:
            if last_discharge < admission_time:
                new_admission_id = admission_id
            if new_admission_id != admission_id:
                new_admission[admission_id] = new_admission_id
            last_discharge = max(last_discharge, discharge_time)
        return dict(new_admission)


class MergeOverlappingAdmissions(ProcessOverlappingAdmissions):
    @staticmethod
    def _tables_map_admission_ids(dataset: Dataset, sub2sup: dict[str, str], report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        c_admission_id = dataset.config.columns.admissions.admission_id

        target_tables = {  # columns that have admission_id as column
            k: v for k, v in tables_dict.items() if k != "admissions" and c_admission_id in v.columns
        }

        tables = dataset.tables
        for table_name, table in target_tables.items():
            n1 = table[c_admission_id].nunique()
            table.loc[:, c_admission_id] = table.loc[:, c_admission_id].map(lambda i: sub2sup.get(i, i))
            n2 = table[c_admission_id].nunique()
            report = report.add(
                table=table_name,
                column=c_admission_id,
                before=n1,
                after=n2,
                value_type="nunique",
                operation="map_admission_id",
            )
            tables = eqx.tree_at(lambda x: getattr(x, table_name), tables, table)

        return eqx.tree_at(lambda x: x.tables, dataset, tables), report

    @staticmethod
    def _admissions_map_admission_ids(
        dataset: Dataset, sub2sup: dict[str, str], report: Report
    ) -> tuple[Dataset, Report]:
        admissions = dataset.tables.admissions

        # Step 1: Map from super-admissions to its sub-admissions.
        sup2sub = defaultdict(list)
        for sub, sup in sub2sup.items():
            sup2sub[sup].append(sub)

        # Step 2: Merge overlapping admissions by extending discharge time to the maximum discharge
        # time of its sub-admissions.
        for super_idx, sub_indices in sup2sub.items():
            current_dischtime = admissions.loc[super_idx, COLUMN.end_time]
            new_dischtime = max(max(admissions.loc[sub_indices, COLUMN.end_time].values), current_dischtime)
            admissions.loc[super_idx, COLUMN.end_time] = new_dischtime

        # Step 3: Remove sub-admissions.
        n1 = len(admissions)
        admissions = admissions.drop(list(sub2sup.keys()), axis="index")
        n2 = len(admissions)
        dataset = eqx.tree_at(lambda x: x.tables.admissions, dataset, admissions)
        report = report.add(
            table="admissions",
            column=COLUMN.admission_id,
            value_type="count",
            operation="merge_overlapping_admissions",
            before=n1,
            after=n2,
        )
        return dataset, report

    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        # Step 1: Collect overlapping admissions
        # NOTE: assumes unique admissions globally. See ISSUE_ADM_UNIQ in dataset.py.
        sub2sup = {
            adm_id: super_adm_id
            for _, subject_admissions in dataset.tables.admissions.groupby(COLUMN.subject_id)
            for adm_id, super_adm_id in cls._collect_overlaps(subject_admissions).items()
        }
        # Step 2: Merge in admissions table
        dataset, report = cls._admissions_map_admission_ids(dataset, sub2sup, report)
        # Step 3: Map admission ids in other tables.
        return cls._tables_map_admission_ids(dataset, sub2sup, report)


class RemoveSubjectsWithOverlappingAdmissions(ProcessOverlappingAdmissions):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        admissions = dataset.tables.admissions
        # Step 1: Collect overlapping admissions
        # Map from sub-admissions to the new super-admissions.
        sub2sup = {
            adm_id: super_adm_id
            for _, subject_admissions in admissions.groupby(COLUMN.subject_id)
            for adm_id, super_adm_id in cls._collect_overlaps(subject_admissions).items()
        }

        # Step 2: Collect subjects with at least one overlapping admission and remove them entirely.
        subject_ids = admissions.loc[sub2sup.keys(), COLUMN.subject_id].unique()
        static = dataset.tables.static
        n1 = len(static)
        static = static.drop(subject_ids, axis="index")
        n2 = len(static)
        report = report.add(
            table="static",
            column=COLUMN.subject_id,
            value_type="count",
            operation="filter_problematic_subjects",
            before=n1,
            after=n2,
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        # Step 4: synchronize subjects
        return cls.synchronize_subjects(dataset, report)


class FilterClampTimestampsToAdmissionInterval(DatasetTransformation):
    @classmethod
    def _filter_timestamped_tables(cls, dataset: Dataset, report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        timestamped_tables_conf = dataset.config.columns.timestamped_tables_config_dict
        timestamped_tables = {name: tables_dict[name] for name in timestamped_tables_conf.keys() if name in tables_dict}

        table_config = dataset.config.columns.admissions
        c_admission_id = table_config.admission_id
        c_dischtime = table_config.end_time
        c_admittime = table_config.start_time
        admissions = dataset.tables.admissions[[c_admittime, c_dischtime]]

        for name, table in timestamped_tables.items():
            c_time = timestamped_tables_conf[name].time
            df = pd.merge(
                table, admissions, how="left", left_on=c_admission_id, right_index=True, suffixes=(None, "_y")
            )
            admittime_col = f"{c_admittime}_y" if c_admittime in table.columns else c_admittime
            dischtime_col = f"{c_dischtime}_y" if c_dischtime in table.columns else c_dischtime

            index = df[df[c_time].between(df[admittime_col], df[dischtime_col])].index
            n1 = len(table)
            table = table.loc[index]
            n2 = len(table)
            report = report.add(table=name, column=c_time, value_type="count", operation="filter", before=n1, after=n2)
            dataset = eqx.tree_at(lambda x: getattr(x.tables, name), dataset, table)

        return dataset, report

    @classmethod
    def _filter_interval_based_tables(cls, dataset: Dataset, report: Report) -> tuple[Dataset, Report]:
        tables_dict = dataset.tables.tables_dict
        interval_based_tables_conf = dataset.config.columns.interval_based_table_config_dict
        interval_based_tables: dict[str, pd.DataFrame] = {
            name: tables_dict[name] for name in interval_based_tables_conf.keys() if name in tables_dict
        }
        table_config = dataset.config.columns.admissions
        c_admission_id = table_config.admission_id
        c_dischtime = table_config.end_time
        c_admittime = table_config.start_time
        admissions = dataset.tables.admissions[[c_admittime, c_dischtime]]

        for name, table in interval_based_tables.items():
            c_start_time = interval_based_tables_conf[name].start_time
            c_end_time = interval_based_tables_conf[name].end_time
            df = pd.merge(
                table, admissions, how="left", left_on=c_admission_id, right_index=True, suffixes=(None, "_y")
            )
            admittime_col = f"{c_admittime}_y" if c_admittime in table.columns else c_admittime
            dischtime_col = f"{c_dischtime}_y" if c_dischtime in table.columns else c_dischtime
            # Step 1: Filter out intervals that are entirely outside admission interval.
            index = df[
                df[c_start_time].between(df[admittime_col], df[dischtime_col])
                | df[c_end_time].between(df[admittime_col], df[dischtime_col])
            ].index
            n1 = len(df)
            df = df.loc[index]
            n2 = len(df)
            report = report.add(
                table=name,
                column=(c_start_time, c_end_time),
                value_type="count",
                operation="filter",
                before=n1,
                after=n2,
            )

            # Step 2: Clamp intervals to admission interval if either side is outside.
            n_to_clamp = np.sum((df[c_start_time] < df[admittime_col]) | (df[c_end_time] > df[dischtime_col]))
            report = report.add(
                table=name,
                column=(c_start_time, c_end_time),
                value_type="count",
                operation="clamp",
                before=None,
                after=n_to_clamp,
            )
            df[c_start_time] = df[c_start_time].clip(lower=df[admittime_col], upper=df[dischtime_col])
            df[c_end_time] = df[c_end_time].clip(lower=df[admittime_col], upper=df[dischtime_col])
            df = df[table.columns]
            dataset = eqx.tree_at(lambda x: getattr(x.tables, name), dataset, df)

        return dataset, report

    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        dataset, report = cls._filter_timestamped_tables(dataset, report)
        return cls._filter_interval_based_tables(dataset, report)


class SelectSubjectsWithObservation(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        c_code = dataset.config.columns.obs.code
        c_admission_id = dataset.config.columns.obs.admission_id
        c_subject = dataset.config.columns.static.subject_id
        obs = dataset.tables.obs
        assert isinstance(obs, pd.DataFrame)

        code = dataset.config.select_subjects_with_observation
        assert isinstance(code, str), "No code provided for filtering subjects"

        admission_ids = obs.loc[obs[c_code] == code, c_admission_id].unique()
        assert len(admission_ids) > 0, f"No observations for code {code}"

        subjects = dataset.tables.admissions.loc[admission_ids, c_subject].unique()
        static = dataset.tables.static
        n1 = len(static)
        static = static[static.index.isin(subjects)]
        n2 = len(static)
        report = report.add(
            table="static",
            column=c_subject,
            value_type="count",
            operation=f"select_subjects(has({code}))",
            before=n1,
            after=n2,
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class FilterInvalidInputRatesSubjects(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        c_rate = dataset.config.columns.icu_inputs.derived_normalized_amount_per_hour
        c_admission_id = dataset.config.columns.admissions.admission_id
        c_subject_id = dataset.config.columns.admissions.subject_id

        icu_inputs = dataset.tables.icu_inputs
        assert isinstance(icu_inputs, pd.DataFrame)
        static = dataset.tables.static
        admissions = dataset.tables.admissions

        nan_input_rates = icu_inputs[icu_inputs[c_rate].isnull()]
        n_nan_inputs = len(nan_input_rates)
        nan_adm_ids = nan_input_rates.loc[:, c_admission_id].unique()
        n_nan_adms = len(nan_adm_ids)

        nan_subject_ids = admissions.loc[admissions.index.isin(nan_adm_ids), c_subject_id].unique()
        n_nan_subjects = len(nan_subject_ids)

        report = report.add(
            table=("icu_inputs", "admissions", "static"),
            column=(c_rate, c_admission_id, c_subject_id),
            value_type="nan_counts",
            before=(n_nan_inputs, n_nan_adms, n_nan_subjects),
            after=None,
            operation="filter_invalid_input_rates_subjects",
        )

        n1 = len(static)
        static = static[~static.index.isin(nan_subject_ids)]
        n2 = len(static)
        report = report.add(
            table="static",
            column=c_subject_id,
            value_type="count",
            before=n1,
            after=n2,
            operation="filter_invalid_input_rates_subjects",
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class FilterSubjectsWithInvalidInputInterval(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        c_admission_id = dataset.config.columns.admissions.admission_id
        c_subject_id = dataset.config.columns.admissions.subject_id
        c_start = dataset.config.columns.icu_inputs.start_time
        c_end = dataset.config.columns.icu_inputs.end_time

        icu_inputs = dataset.tables.icu_inputs
        assert isinstance(icu_inputs, pd.DataFrame)
        static = dataset.tables.static
        admissions = dataset.tables.admissions

        invalid_input_interval = icu_inputs[icu_inputs[c_start] > icu_inputs[c_end]]
        n_invalid_inputs = len(invalid_input_interval)
        invalidated_adm_ids = invalid_input_interval.loc[:, c_admission_id].unique()
        n_invalidated_adms = len(invalidated_adm_ids)

        invalidated_subject_ids = admissions.loc[admissions.index.isin(invalidated_adm_ids), c_subject_id].unique()
        n_nan_subjects = len(invalidated_subject_ids)

        report = report.add(
            table=("icu_inputs", "admissions", "static"),
            column=(c_start, c_admission_id, c_subject_id),
            value_type="invalid_counts",
            before=(n_invalid_inputs, n_invalidated_adms, n_nan_subjects),
            after=None,
            operation="filter_invalid_input_intervals_subjects",
        )

        n1 = len(static)
        static = static[~static.index.isin(invalidated_subject_ids)]
        n2 = len(static)
        report = report.add(
            table="static",
            column=c_subject_id,
            value_type="count",
            before=n1,
            after=n2,
            operation="filter_invalid_input_intervals_subjects",
        )
        dataset = eqx.tree_at(lambda x: x.tables.static, dataset, static)
        return cls.synchronize_subjects(dataset, report)


class ICUInputRateUnitConversion(DatasetTransformation):
    @classmethod
    def apply(cls, dataset: Dataset, schemes_context: CodingSchemesManager, report: Report) -> tuple[Dataset, Report]:
        ds_config = dataset.config
        tables_config = ds_config.columns
        table_config = tables_config.icu_inputs
        c_code = table_config.code
        c_amount = table_config.amount
        c_start_time = table_config.start_time
        c_end_time = table_config.end_time
        c_amount_unit = table_config.amount_unit
        c_normalized_amount = table_config.derived_normalized_amount
        c_normalized_amount_per_hour = table_config.derived_normalized_amount_per_hour
        c_universal_unit = table_config.derived_universal_unit
        c_normalization_factor = table_config.derived_unit_normalization_factor
        icu_inputs = dataset.tables.icu_inputs
        assert isinstance(icu_inputs, pd.DataFrame)

        scheme = dataset.scheme_proxy(schemes_context).icu_inputs
        assert isinstance(scheme, CodingSchemeWithUOM), f"Expected CodingSchemeWithUOM but got {type(scheme)}"
        _derived_columns = [c_normalized_amount, c_normalized_amount_per_hour, c_universal_unit, c_normalization_factor]

        assert (c in icu_inputs.columns for c in [c_code, c_amount, c_amount_unit]), (
            f"Some columns in: {c_code}, {c_amount}, {c_amount_unit}, not found in icu_inputs table"
        )
        assert all(c not in icu_inputs.columns for c in _derived_columns), (
            f"Some of these columns [{', '.join(_derived_columns)}] already exists in icu_inputs table"
        )
        df = icu_inputs.iloc[:, :]
        df[c_universal_unit] = df[c_code].map(lambda c: scheme.universal_unit[c])
        df[c_normalization_factor] = [
            scheme.uom_normalization_factor[code][unit] for code, unit in zip(df[c_code], df[c_amount_unit].str.lower())
        ]

        delta_hours = (df[c_end_time] - df[c_start_time]).dt.total_seconds() * SECONDS_TO_HOURS_SCALER
        df[c_normalized_amount] = df[c_amount] * df[c_normalization_factor]
        df[c_normalized_amount_per_hour] = df[c_normalized_amount] / delta_hours
        df = df[icu_inputs.columns.tolist() + _derived_columns]
        dataset = eqx.tree_at(lambda x: x.tables.icu_inputs, dataset, df)
        report = report.add(
            table="icu_inputs",
            column=None,
            value_type="columns",
            operation="new_columns",
            before=icu_inputs.columns.tolist(),
            after=df.columns.tolist(),
        )

        return dataset, report



from typing import Final

import equinox as eqx
import sqlalchemy

from ..coding_scheme import CodingSchemesManager
from ..dataset import AbstractDatasetPipeline, Dataset, DatasetColumns, DatasetConfig, DatasetSchemeConfig
from ..transformations import (
    CastTimestamps,
    FilterAdmissionsWithNoObservables, FilterClampTimestampsToAdmissionInterval,
    FilterInvalidInputRatesSubjects,
    FilterShortAdmissions,
    FilterSubjectsNegativeAdmissionLengths,
    FilterSubjectsWithInvalidInputInterval,
    FilterUnsupportedCodes,
    ICUInputRateUnitConversion,
    MergeOverlappingAdmissions,
    SelectSubjectsWithObservation,
    SetAdmissionRelativeTimes,
    SetIndex,
)
from ..tvx_ehr import (
    AbstractTVxPipeline,
    DatasetNumericalProcessorsConfig,
    DemographicVectorConfig,
    LeadingObservableExtractorConfig,
    TVxEHRConfig,
    TVxEHRSampleConfig,
    TVxEHRSchemeConfig,
    TVxEHRSplitsConfig,
)
from ..tvx_transformations import (
    InputScaler,
    InterventionSegmentation,
    LeadingObservableExtraction,
    ObsAdaptiveScaler,
    ObsIQROutlierRemover,
    ObsTimeBinning,
    RandomSplits,
    SampleSubjects,
    TVxConcepts,
)
from .mimic import load_mimic, MIMICDatasetAuxiliaryResources, MIMICDatasetSchemeSuffixes, ScopedSchemeNames
from .mimic_sql import SQLMIMICTablesResources


OBSERVABLE_AKI_TARGET_CODE: Final[str] = "renal_aki.aki_binary"


def default_suffixes() -> MIMICDatasetSchemeSuffixes:
    return MIMICDatasetSchemeSuffixes(
        ethnicity="ethnicity",
        gender="gender",
        obs="obs",
        dx_discharge="dx_discharge",
        hosp_procedures="hosp_procedures",
        icu_procedures="icu_procedures",
        icu_inputs="icu_inputs",
    )


def default_auxiliary_resources() -> MIMICDatasetAuxiliaryResources:
    return MIMICDatasetAuxiliaryResources.make_resources(
        name_prefix="mimiciv_aki",
        resources_root="aki_study",
        suffixes=default_suffixes(),
        icu_inputs_uom_normalization=("uom_normalization", "icu_inputs.csv"),
        icu_inputs_aggregation_column="aggregation",
    )


def default_scoped_names() -> ScopedSchemeNames:
    return default_auxiliary_resources().scoped_names


def default_dataset_schemes_config(scoped_names: ScopedSchemeNames) -> DatasetSchemeConfig:
    return DatasetSchemeConfig(
        ethnicity=scoped_names.ethnicity,
        gender=scoped_names.gender,
        dx_discharge=scoped_names.dx_discharge,
        obs=scoped_names.obs,
        icu_procedures=scoped_names.icu_procedures,
        hosp_procedures=scoped_names.hosp_procedures,
        icu_inputs=scoped_names.icu_inputs,
    )


def default_dataset_config(scoped_names: ScopedSchemeNames) -> DatasetConfig:
    return DatasetConfig(
        scheme=default_dataset_schemes_config(scoped_names),
        columns=DatasetColumns(),
        select_subjects_with_observation=OBSERVABLE_AKI_TARGET_CODE,
        admission_minimum_los=12.0 / 24.0,  # 12 hours.
    )


def default_dataset_pipeline() -> AbstractDatasetPipeline:
    pipeline = [
        SetIndex(),
        SelectSubjectsWithObservation(),
        CastTimestamps(),
        MergeOverlappingAdmissions(),
        FilterSubjectsNegativeAdmissionLengths(),
        FilterAdmissionsWithNoObservables(),
        FilterShortAdmissions(),
        FilterClampTimestampsToAdmissionInterval(),
        FilterUnsupportedCodes(),
        ICUInputRateUnitConversion(),
        FilterInvalidInputRatesSubjects(),
        FilterSubjectsWithInvalidInputInterval(),
        SetAdmissionRelativeTimes(),
    ]
    return AbstractDatasetPipeline(transformations=pipeline)


def default_tvx_schemes_config(config: DatasetSchemeConfig, scoped_names: ScopedSchemeNames) -> TVxEHRSchemeConfig:
    names = scoped_names.mapped
    return TVxEHRSchemeConfig(
        gender=config.gender,
        ethnicity=names.ethnicity,
        dx_discharge="dx_ccs",
        obs=config.obs,
        icu_inputs=names.icu_inputs,
        icu_procedures=names.icu_procedures,
        hosp_procedures=names.hosp_procedures,
        outcome="dx_flat_ccs_v1",
    )


def default_tvx_ehr_config(scoped_names: ScopedSchemeNames = default_scoped_names()) -> TVxEHRConfig:
    scheme = default_tvx_schemes_config(default_dataset_schemes_config(scoped_names), scoped_names)
    return TVxEHRConfig(
        scheme=scheme,
        demographic=DemographicVectorConfig(age=True, gender=True, ethnicity=True),
        leading_observable=LeadingObservableExtractorConfig(
            observable_code=OBSERVABLE_AKI_TARGET_CODE,
            scheme=scheme.obs,
            leading_hours=[6.0, 12.0, 24.0, 48.0, 72.0],  # hours
            entry_neglect_window=6.0,  # hours
            minimum_acquisitions=2,  # number of observable acquisitions.
            recovery_window=12.0,
        ),  # hours
        sample=TVxEHRSampleConfig(n_subjects=6000, seed=0, offset=0),  # no subsetting now
        splits=TVxEHRSplitsConfig(
            split_quantiles=[0.6, 0.7, 0.8], seed=0, discount_first_admission=False, balance="admissions"
        ),
        numerical_processors=DatasetNumericalProcessorsConfig(),
        interventions=True,
        observables=True,
        time_binning=None,
        interventions_segmentation=True,
    )


def default_tb_tvx_ehr_config(scoped_names: ScopedSchemeNames = default_scoped_names()) -> TVxEHRConfig:
    config = default_tvx_ehr_config(scoped_names)
    return eqx.tree_at(lambda x: x.time_binning, config, 6.0, is_leaf=lambda x: x is None)


def default_tvx_ehr_pipeline() -> AbstractTVxPipeline:
    pipeline = [
        SampleSubjects(),
        RandomSplits(),
        ObsIQROutlierRemover(),
        ObsAdaptiveScaler(),
        InputScaler(),
        TVxConcepts(),
        ObsTimeBinning(),
        LeadingObservableExtraction(),
        InterventionSegmentation(),
    ]
    return AbstractTVxPipeline(transformations=pipeline)


def mimiciv_from_env_sql(
    dataset_tables_resources: SQLMIMICTablesResources = SQLMIMICTablesResources(),
    aux_resources: MIMICDatasetAuxiliaryResources = default_auxiliary_resources(),
) -> tuple[Dataset, CodingSchemesManager]:
    engine = sqlalchemy.create_engine(dataset_tables_resources.url())
    return load_mimic(
        config=default_dataset_config(aux_resources.scoped_names),
        tables=dataset_tables_resources,
        aux=aux_resources,
        data_connection=engine,
    )
